Attribute VB_Name = "basMagvar"
' Module to calculate magnetic variation and field given position,
'                altitude, and date
'  Implements the NIMA (formerly DMA) WMM and IGRF models
'
'     http://www.nima.mil/GandG/ngdc-wmm2000.html
'     For WMM2000 coefficients:
'     ftp://ftp.ngdc.noaa.gov/Solid_Earth/Mainfld_Mag/DoD_Model/wmm.cof
'     For IGRF/DGRF coefficients:
'     http://swdcdb.kugi.kyoto-u.ac.jp/igrf/coef/igrfall.d
'
'  Copyright (C) 2000  Edward A Williams <Ed_Williams@compuserve.com>
'
'   The routine uses a spherical harmonic expansion of the magnetic
'  potential up to twelfth order, together with its time variation, as
'  described in Chapter 4 of "Geomagnetism, Vol 1, Ed. J.A.Jacobs,
'  Academic Press (London 1987)". The program first converts geodetic
'  coordinates (lat/long on elliptic earth and altitude) to spherical
'  geocentric (spherical lat/long and radius) coordinates. Using this,
'  the spherical (B_r, B_theta, B_phi) magnetic field components are
'  computed from the model. These are finally referred to surface (X, Y,
'  Z) coordinates.
'
'    Fields are accurate to better than 200nT, variation and dip to
'  better than 0.5 degrees, with the exception of the declination near
'  the magnetic poles (where it is ill-defined) where the error may reach
'  4 degrees or more.
'
'    Variation is undefined at both the geographic and
'  magnetic poles, even though the field itself is well-behaved. To
'  avoid the routine blowing up, latitude entries corresponding to
'  the geographic poles are slightly offset. At the magnetic poles,
'  the routine returns zero variation.
'
'  HISTORY
'  Adapted from EAW Excel 3.0 version 3/27/94 EAW
'  Recoded in C++ by Starry Chan
'  WMM95 added and rearranged in ANSI-C EAW 7/9/95
'  Put shell around program and made Borland & GCC compatible EAW 11/22/95
'  IGRF95 added 2/96 EAW
'  WMM2000 IGR2000 added 2/00 EAW
'  Released under GPL  3/26/00 EAW
'  Adaptions and modifications for the SimGear project  3/27/2000 CLO
'  Removed all pow() calls and made static roots[][] arrays to
'  save many sqrt() calls on subsequent invocations
'  3/28/2000  Norman Vine -- nhv@yahoo.com
'  Put in some bullet-proofing to handle magnetic and geographic poles.
'  3/28/2000 EAW
'  Added missing comment close, the lack of which caused the altitude
'  correction to be omitted.
'  01/31/01 Jim Seymour (jseymour@LinxNet.com)
'  08/20/03 Translated into Visual Basic - Ian Billings
'  This version assumes IGRF2000, H=0 and a date of 1 September 2003
'
'
'  This program is free software; you can redistribute it and/or
'  modify it under the terms of the GNU General Public License as
'  published by the Free Software Foundation; either version 2 of the
'  License, or (at your option) any later version.
'
'  This program is distributed in the hope that it will be useful, but
'  WITHOUT ANY WARRANTY; without even the implied warranty of
'  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
'  General Public License for more details.
'
'  You should have received a copy of the GNU General Public License
'  along with this program; if not, write to the Free Software
'  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 


Option Explicit

Private gnm_igrf2000(12, 12)
Private hnm_igrf2000(12, 12)
Private gtnm_igrf2000(12, 12)
Private htnm_igrf2000(12, 12)

Const pi = 3.14159265358979

Private Sub Main()

Dim Latitude As String
Dim Longitude As String
Dim Variation As Double
Dim Message, Title, Default
Dim strType As String
Dim Lat As Double
Dim Lon As Double
Dim EW As String

Message = "Enter a latitude Format [NS]DDMM.MMMM eg. Sxxxx.x"    ' Set prompt.
Title = "Latitude"    ' Set title.
Default = "S2608.1"    ' Set default.
Latitude = InputBox(Message, Title, Default)
strType = "Lat"
If Not CheckLatLon(Latitude, strType) Then Exit Sub
Lat = ConvertLatLon(Latitude, strType)
If Abs(Lat) > pi / 2 Then
MsgBox Latitude & " is not valid.", vbCritical, "Only values from 0 - 90 are valid"
Exit Sub
End If

Message = "Enter a longitude Format [EW]DDDMM.MMMM eg. Exxxxx.x" ' Set prompt."
Title = "Longitude"    ' Set title.
Default = "E02814.1"    ' Set default.
Longitude = InputBox(Message, Title, Default)
strType = "Lon"
If Not CheckLatLon(Longitude, strType) Then Exit Sub
Lon = ConvertLatLon(Longitude, strType)
If Abs(Lon) > pi Then
MsgBox Longitude & " is not valid.", vbCritical, "Only values from 0 - 180 are valid"
Exit Sub
End If

Variation = MagVar(Lat, Lon)

If Variation <= 0 Then
EW = "W"
Else
EW = "E"
End If

MsgBox "The variation is " & EW & Abs(Variation), vbInformation, Latitude & " / " & Longitude


End Sub

Public Function MagVar(Lat As Double, Lon As Double) As Integer
On Error GoTo Err

 ' return variation (in radians) given geodetic latitude (radians), longitude
 ' (radians) ,height (km) and (Julian) date
 ' model=1 is IGRF90, 2 is WMM85, 3 is WMM90, 4 is WMM95,
 ' 5 is IGRF95, 6 is WMM2000, 7 is IGRF2000
 ' N and E lat and long are positive, S and W negative
 ' output field B_r,B_th,B_phi,B_x,B_y,B_z */
 
Const A = 6378.16       ' major radius (km) IAU66 ellipsoid
Const F = 1# / 298.25     ' inverse flattening IAU66 ellipsoid
Const B = 6378.16 * (1# - 1# / 298.25)   ' minor radius b=a*(1-f)
Const r_0 = 6371.2      '"mean radius" for spherical harmonic expansion
Const H = 0             'Height is always sea level
Const Model = 7         'model is always IGRF2000 model

    Dim n As Integer, m As Integer
    Dim yearfrac As Double, sr As Double, r As Double, theta As Double, C As Double
    Dim s As Double, PSI As Double, fn As Double, fn_0 As Double, B_r As Double
    Dim B_theta As Double, B_phi As Double, X As Double, Y As Double, Z As Double
    Dim sinpsi As Double, cospsi As Double, inv_s As Double
    Dim Var As Double
    Static been_here As Integer
    Dim sinlat As Double
    Dim coslat As Double
    Dim P(12, 12)
    Dim DP(12, 12)
    Dim GNM(12, 12)
    Dim hnm(12, 12)
    Dim sm(12)
    Dim cm(12)
    Static root(12)
    Static roots(13, 13, 2)
    Dim field(5)
    Dim nmax As Integer
    Dim dat As Long
    
    dat = yymmdd_to_julian_days(3, 9, 1) '1 September 2003
    
    nmax = 12
    been_here = 0
    sinlat = Sin(Lat)
    coslat = Cos(Lat)

    IGRF_arrays 'Function to populate arrays with IGRF2000 data

    'convert to geocentric
    sr = Sqr(A * A * coslat * coslat + B * B * sinlat * sinlat)
    'sr is effective radius
    theta = atan2((coslat * (H * sr + A * A)), (sinlat * (H * sr + B * B)))
          
    'theta is geocentric co-latitude */

    r = H * H + 2# * H * sr + _
    (A * A * A * A - (A * A * A * A - B * B * B * B) * sinlat * sinlat) / _
    (A * A - (A * A - B * B) * sinlat * sinlat)

    r = Sqr(r)

    'r is geocentric radial distance
    C = Cos(theta)
    s = Sin(theta)
    'protect against zero divide at geographic poles */
    inv_s = 1# / (s + (s = 0) * 0.00000001)

    'zero out arrays
  Do While m < 13
    Do While n < 12
    P(n, m) = 0
    n = n + 1
    Loop
  n = 0
  m = m + 1
  Loop
  
  'Diagonal elements
    P(0, 0) = 1
    P(1, 1) = s
    DP(0, 0) = 0
    DP(1, 1) = C
    P(1, 0) = C
    DP(1, 0) = -s
        
'These values will not change for subsequent function calls
If been_here = 0 Then
    For n = 2 To nmax
    root(n) = Sqr((2# * n - 1) / (2# * n))
    Next n
      
    For m = 0 To nmax
        Dim mm As Double
        mm = m * m
        If m + 1 > 2 Then n = m + 1
        If m + 1 < 2 Then n = 2
        If m + 1 = 2 Then n = 2
        For n = n To nmax
        roots(m, n, 0) = Sqr((n - 1) * (n - 1) - mm)
        roots(m, n, 1) = 1# / Sqr(n * n - mm)
        Next n
    Next m
been_here = 1
End If

'Double root = sqrt((2.0*n-1) / (2.0*n)); */
For n = 2 To nmax
    P(n, n) = P(n - 1, n - 1) * s * root(n)
    DP(n, n) = (DP(n - 1, n - 1) * s + P(n - 1, n - 1) * C) * root(n)
Next n

'Lower triangle */
For m = 0 To nmax
        If m + 1 > 2 Then n = m + 1
        If m + 1 < 2 Then n = 2
        If m + 1 = 2 Then n = 2
        For n = n To nmax
        P(n, m) = (P(n - 1, m) * C * (2# * n - 1) - P(n - 2, m) * roots(m, n, 0)) * roots(m, n, 1)
        DP(n, m) = ((DP(n - 1, m) * C - P(n - 1, m) * s) * (2# * n - 1) - DP(n - 2, m) * roots(m, n, 0)) * roots(m, n, 1)
        Next n
Next m

 'IGRF2000 */
      yearfrac = (dat - yymmdd_to_julian_days(0, 1, 1)) / 365.25
For n = 1 To nmax
      For m = 0 To nmax
      GNM(n, m) = gnm_igrf2000(n, m) + yearfrac * gtnm_igrf2000(n, m)
      hnm(n, m) = hnm_igrf2000(n, m) + yearfrac * htnm_igrf2000(n, m)
      Next m
Next n

'Compute sm (sin(m lon) and cm (cos(m lon)) */
    For m = 0 To nmax
      sm(m) = Sin(m * Lon)
      cm(m) = Cos(m * Lon)
    Next m
    
'Compute B fields */
    B_r = 0#
    B_theta = 0#
    B_phi = 0#
    fn_0 = r_0 / r
    fn = fn_0 * fn_0
    
    For n = 1 To nmax
      Dim c1_n, c2_n, c3_n
      c1_n = 0
      c2_n = 0
      c3_n = 0
        For m = 0 To nmax
        Dim tmp
        tmp = (GNM(n, m) * cm(m) + hnm(n, m) * sm(m))
        c1_n = c1_n + (tmp * P(n, m))
        c2_n = c2_n + (tmp * DP(n, m))
        c3_n = c3_n + (m * (GNM(n, m) * sm(m) - hnm(n, m) * cm(m)) * P(n, m))
        Next m
     'fn=pow(r_0/r,n+2.0);   */
      fn = fn_0 * fn_0
      B_r = B_r + ((n + 1) * c1_n * fn)
      B_theta = B_theta - (c2_n * fn)
      B_phi = B_phi + (c3_n * fn * inv_s)
   Next n
   
' Find geodetic field components: */
    PSI = theta - (pi / 2# - Lat)
    sinpsi = Sin(PSI)
    cospsi = Cos(PSI)
    X = -B_theta * cospsi - B_r * sinpsi
    Y = B_phi
    Z = B_theta * sinpsi - B_r * cospsi

    field(0) = B_r
    field(1) = B_theta
    field(2) = B_phi
    field(3) = X
    field(4) = Y
    field(5) = Z        ' output fields */
    '/* find variation in radians */
    '/* return zero variation at magnetic pole X=Y=0. */
    '/* E is positive */
    'return (X != 0. || Y != 0.) ? atan2(Y, X) : (double) 0
    Var = atan2(Y, X)
    Var = rad_to_deg(Var)
    Var = Round(Var, 0)
    MagVar = Var
    
ErrExit:
Exit Function

Err:
MsgBox Error$, vbCritical, "Mag variation calculation error"
Resume ErrExit
   
End Function
Sub IGRF_arrays()
Dim ary
Dim P, q, r

'static const double gnm_igrf2000[13][13]
    ary = Array(0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    -29615#, -1728#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    -2267#, 3072#, 1672#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    1341#, -2290#, 1253#, 715#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    935#, 787#, 251#, -405#, 110#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    -217#, 351#, 222#, -131#, -169#, -12#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    72#, 68#, 74#, -161#, -5#, 17#, -91#, 0#, 0#, 0#, 0#, 0#, 0#, _
    79#, -74#, 0#, 33#, 9#, 7#, 8#, -2#, 0#, 0#, 0#, 0#, 0#, _
    25#, 6#, -9#, -8#, -17#, 9#, 7#, -8#, -7#, 0#, 0#, 0#, 0#, _
    5#, 9#, 3#, -8#, 6#, -9#, -2#, 9#, -4#, -8#, 0#, 0#, 0#, _
    -2#, -6#, 2#, -3#, 0#, 4#, 1#, 2#, 4#, 0#, -1#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#)

r = 0
For P = 0 To 12
    For q = 0 To 12
    gnm_igrf2000(P, q) = ary(r)
    r = r + 1
    Next q
Next P

'static const double hnm_igrf2000[13][13]
    ary = Array(0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 5186#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -2478#, -458#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -227#, 296#, -492#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 272#, -232#, 119#, -304#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 44#, 172#, -134#, -40#, 107#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -17#, 64#, 65#, -61#, 1#, 44#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -65#, -24#, 6#, 24#, 15#, -25#, -6#, 0#, 0#, 0#, 0#, 0#, _
    0#, 12#, -22#, 8#, -21#, 15#, 9#, -16#, -3#, 0#, 0#, 0#, 0#, _
    0#, -20#, 13#, 12#, -6#, -8#, 9#, 4#, -8#, 5#, 0#, 0#, 0#, _
    0#, 1#, 0#, 4#, 5#, -6#, -1#, -3#, 0#, -2#, -8#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#)
r = 0
For P = 0 To 12
    For q = 0 To 12
    hnm_igrf2000(P, q) = ary(r)
    r = r + 1
    Next q
Next P

'static const double gtnm_igrf2000[13][13]
    ary = Array(0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    14.6, 10.7, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    -12.4, 1.1, -1.1, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0.7, -5.4, 0.9, -7.7, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    -1.3, 1.6, -7.3, 2.9, -3.2, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -0.7, -2.1, -2.8, -0.8, 2.5, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    1#, -0.4, 0.9, 2#, -0.6, -0.3, 1.2, 0#, 0#, 0#, 0#, 0#, 0#, _
    -0.4, -0.4, -0.3, 1.1, 1.1, -0.2, 0.6, -0.9, 0#, 0#, 0#, 0#, 0#, _
    -0.3, 0.2, -0.3, 0.4, -1#, 0.3, -0.5, -0.7, -0.4, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#)
r = 0
For P = 0 To 12
    For q = 0 To 12
    gtnm_igrf2000(P, q) = ary(r)
    r = r + 1
    Next q
Next P

'static const double htnm_igrf2000[13][13]
    ary = Array(0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -22.5, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -20.6, -9.6, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 6#, -0.1, -14.2, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 2.1, 1.3, 5#, 0.3, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -0.1, 0.6, 1.7, 1.9, 0.1, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, -0.2, -1.4, 0#, -0.8, 0#, 0.9, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 1.1, 0#, 0.3, -0.1, -0.6, -0.7, 0.2, 0#, 0#, 0#, 0#, 0#, _
    0#, 0.1, 0#, 0#, 0.3, 0.6, -0.4, 0.3, 0.7, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, _
    0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#, 0#)
r = 0
For P = 0 To 12
    For q = 0 To 12
    htnm_igrf2000(P, q) = ary(r)
    r = r + 1
    Next q
Next P

End Sub
Private Function yymmdd_to_julian_days(yy As Integer, mm As Integer, dd As Integer) As Long
'Convert date to Julian day    1950-2049
Dim jd As Long

  If yy < 50 Then yy = (2000 + yy) Else yy = (1900 + yy)
  jd = dd - 32075 + 1461 * (yy + 4800 + (mm - 14) / 12) / 4
  jd = jd + 367 * (mm - 2 - (mm - 14) / 12 * 12) / 12
  jd = jd - 3 * ((yy + 4900 + (mm - 14) / 12) / 100) / 4

  yymmdd_to_julian_days = jd
  
End Function

Private Function deg_to_rad(Deg As Double) As Double
'Convert degrees to radians

deg_to_rad = Deg * pi / 180

End Function

Private Function rad_to_deg(rad As Double) As Double
'Convert radians to degrees
rad_to_deg = rad * 180 / pi
End Function
Private Function atan2(Y As Double, X As Double) As Double

If X > 0 Then
atan2 = Atn(Y / X)
ElseIf X < 0 And Y >= 0 Then
atan2 = Atn(Y / X) + pi
ElseIf X = 0 And Y > 0 Then
atan2 = pi / 2
ElseIf X < 0 And Y < 0 Then
atan2 = Atn(Y / X) - pi
ElseIf X = 0 And Y < 0 Then
atan2 = pi / 2
End If

End Function
Private Function CheckLatLon(strLatLon As String, strType As String) As Boolean
On Error GoTo Err

If strType = "Lat" Then
GoTo Latitude
ElseIf strType = "Lon" Then
GoTo Longitude
End If

Latitude:
If Left(strLatLon, 1) <> "N" And Left(strLatLon, 1) <> "S" Then
CheckLatLon = False
GoTo LatError
End If

CheckLatLon = True
Exit Function

LatError:
MsgBox "The format of the latitude entered is incorrect.", vbCritical, "Latitiude format incorrect"
CheckLatLon = False
Exit Function

Longitude:
If Left(strLatLon, 1) <> "W" And Left(strLatLon, 1) <> "E" Then
CheckLatLon = False
GoTo LonError
End If

CheckLatLon = True
Exit Function

LonError:
MsgBox "The format of the longitude entered is incorrect.", vbCritical, "Longitude format incorrect"
CheckLatLon = False
Exit Function

ErrExit:
Exit Function

Err:
MsgBox Error$, vbCritical, "CheckLatLon"
Resume ErrExit

End Function


Private Function ConvertLatLon(strLatLon As String, strType As String) As Double

Dim Deg As Integer
Dim Min As Double

If strType = "Lat" Then
GoTo Latitude
ElseIf strType = "Lon" Then
GoTo Longitude
End If

Latitude:
Deg = Mid(strLatLon, 2, 2)
Min = Mid(strLatLon, 4, 7)

ConvertLatLon = (Deg + (Min / 60)) * (pi / 180)
' Error message if ConvertLatLon > pi/2  - invalid latitude
'Error message placed in calling routine

If Left(strLatLon, 1) = "S" Then
ConvertLatLon = -(ConvertLatLon)
End If
Exit Function

Longitude:
Deg = Mid(strLatLon, 2, 3)
Min = Mid(strLatLon, 5, 2)
' Error message if ConvertLatLon > pi - invalid longitude
'Error message placed in calling routine

ConvertLatLon = (Deg + (Min / 60)) * (pi / 180)
If Left(strLatLon, 1) = "W" Then
ConvertLatLon = -(ConvertLatLon)
End If

End Function


